
!*********************************************************************************************************** 
module omgSrcTable_mod

    use dal
    use cal
    use param  
    use sasdal
    
    use OmAllocation_mod         ! Functions for allocating memory to arrays.
    use OmgSourceparameter_mod             ! Source definitions, etc
    use omconstants_mod
    use OmPhotometryDefs_mod
    use omutility_mod   
    use errorHandling

    implicit none 
 
  private
  public CreateGSourceListFile

contains 
  


!*********************************************
! Subroutine CreateSourceList
!*********************************************
subroutine CreateGSourceListFile(outset,         &
                            spectrumFile,   &
                            oswDataSet,     &

                            nSources,       &
                            sourceDataIn,     &
                            plateScale,     &
                            modeBackground, &
                            nSigma,         &
                            exposure,       & 
                            windowX0,       &
                            windowY0,       & 
                            windowDX,       &
                            windowDY) 
    character(*), intent(in) ::  outset, spectrumFile

    type(DataSetT), intent(in)  :: oswDataSet   ! their associated handles
    integer(int32), intent(in) :: nSources 
    type(gsourceParameters), pointer :: sourceDataIn(:)

    real(double), intent(in) :: plateScale
    real(single), intent(in) :: modeBackground
    real(single), intent(in) :: nSigma
    real(single), intent(in) :: exposure
    integer(int32), intent(in)  :: windowX0     ! WindowX0.
    integer(int32), intent(in)  :: windowY0     ! WindowY0.
    integer(int32), intent(in)  :: windowDX     ! WindowDX.
    integer(int32), intent(in)  :: windowDY     ! WindowDY.
    real(single) :: apertureRadius = 12.0      

    type(DataSetT) :: listDataSet 
    type(TableT) :: listTable
    type(ColumnT)        :: pcolumn                 ! pointer to a column

    integer (int32), pointer :: srcnum(:)  ! the source number
   
   
    integer(int32) :: i
    real (single) :: detectionRate

    type(gsourceParameters), pointer :: sourceData(:) => null()

    if (len(spectrumFile) .lt. 1) then
    end if

    apertureRadius = plateScale

   
    apertureRadius = 12.0




    if (associated(sourceDataIn)) then
        sourceData => sourceDataIn
    else
       allocate(sourceData(1 : 1))
    end if

    ! Write out the source list to the oswDataSet
    write(errstr,'(a)') "Creating source list for omgrism"
    call message(AppMsg, Verbose)

    !**********************************
    ! Create the output source list
    !**********************************


    listDataSet = dataSet(outset, create)

    listTable = addTable(listDataSet, "SRCLIST", nSources)


    ! copy the keywords from the image header
    call copyAllAttributes(attributable(oswDataSet), attributable(listDataSet))

    !*****************************
    ! Add in detection count rate
    ! which will be stored in the source list file
    ! and used by ommag to calculate the
    ! background limiting  magnitude
    !************************************

    if (modeBackground .gt. 0.0) then  
        detectionRate = nSigma * sqrt(modeBackground * PI * apertureRadius * apertureRadius ) / exposure
    else
        detectionRate = 0.0
    end if 

    write(errstr,'(a, 1pg9.3)') "Detection Count Rate: ", detectionRate
    call message(AppMsg, Verbose)

    call setAttribute(listDataSet, "CRLIMIT", detectionRate, "counts/sec", "Detection limit count rate")

    call setAttribute(listDataSet, "WINDOWX0", windowX0, "pixels", "Window x-origin")
    call setAttribute(listDataSet, "WINDOWDX", windowDX, "pixels", "Window dx")
    call setAttribute(listDataSet, "WINDOWY0", windowY0, "pixels", "Window y-origin")
    call setAttribute(listDataSet, "WINDOWDY", windowDY, "pixels", "Window dy")
   
    
  

   
    ! this is an osw source list

    call setAttribute(listDataSet, "CONTENT", "OM OSW SOURCE LIST", &
                  "File contents")
    call setAttribute(listDataSet, "HDUCLASS", "OGIP", &
                  "Format conforms to OGIP/GSFC conventions")
    call setAttribute(listDataSet, "HDUCLAS1", "SRCLIST", &
                  "File contains a source list")

 
       

    !********************************
    ! now add the columns one by one
    ! starting with the source number
    !*********************************
    pcolumn = addColumn(listTable, "SRCNUM", Integer32, " ")
    srcnum => int32data(pcolumn)

    forall (i = 0 : nSources - 1) srcnum(i) = i + 1
   
    call release(pcolumn)


   
    ! Add the source positions



    call AddReal32Column(listTable, sourceData%xPos,  nSources, &
                         "XPOS", "pixels", "Centroided x-pixel position", nullValue)


  
    call AddReal32Column(listTable, sourceData%yPos,  nSources, &
                         "YPOS", "pixels", "Centroided y-pixel position", nullValue)

  
  


    call AddReal32Column(listTable, sourceData%poserr,  nSources, &
                         "POSERR", "pixels", "Centroided pixel position error", nullValue)
    
    call AddInt32Column(listTable, sourceData%noSourcePixels,  nSources, &
                         "NO PIXELS", " ", "Number source pixels", -1)
    


  
    ! Add the source shape parameters 
    ! Convert sigmas to arc secs

!    sourceData%semiA = sourceData%semiA * 2.354 * plateScale * real(binX) / 3.0
!    sourceData%semiB = sourceData%semiB * 2.354 * plateScale * real(binY) / 3.0
  

    ! NB the full-width-half-maximum = 2.354 * sigma (gaussian sigma)
    call AddReal32Column(listTable,  sourceData%semiA, nSources, &
                 "FWHM_MAJ", "arcsecs", "FWHM of semi-major axis", nullValue)
 
    call AddReal32Column(listTable,  sourceData%semiB, nSources, &
                 "FWHM_MIN", "arcsecs", "FWHM of semi-minor axis", nullValue)


 
    call AddReal32Column(listTable,  sourceData%theta, nSources, &
                 "PA", "degrees", "PA of semi-major axis", nullValue)
 
    ! Add the quality flags
    call AddInt16Column(listTable,  sourceData%qualityFlag, nSources, &
                  "QFLAG", "none", "0=0 bad pixels, 1=1 or more bad pixels)")   ! Can be either 0 or 1
    call AddInt8Column(listTable,  sourceData%confusionFlag, nSources, &
                             "CFLAG", "none", "0=isolated")   ! 
    call AddInt8Column(listTable, sourceData%extensionFlag, nSources, &
              "EFLAG", "none", "0=point, 1=extended source")    ! Can be either 0 (point-source) or 1 (extended)


 
    call AddInt32Column(listTable, sourceData%spectrumNo, nSources, &
              "SPECTR_ID", "none", "spectrum identification number       ", -1)    ! Can be either 0 (point-source) or 1 (extended)
 
   call AddInt32Column(listTable, sourceData%origin, nSources, &
              "REL2SRCNUM", "none", "origin", -1)    ! Can be either 0 (point-source) or 1 (extended)
 

    
    !************************
    ! release the memory
    !************************
    call release(listTable)
    call release(listDataSet)
    if (associated(sourceData) .and. .not. associated(sourceDataIn)) deallocate(sourceData) 

end subroutine CreateGSourceListFile




end module omgSrcTable_mod








