! -*-f90-*-
! Originator:cbrindle
! Adapted:   Jorgo Bakker (jbakker@rssd.esa.int)
! Info:rquality
!   Interactive program for selecting sources from an image
!   for further analysis
! Note:
!  * Should always call "omsource_init()" first, and "omsource_finish()" last.
!  * In comments, I usually talk about methods. A method may be either a
!    function OR subroutine within a module
!  * Any variable starting with 'my' is global to this module
   

module omsource_old
   use dal
   use cal
   use sasdal
   use oal
   use param
   use sla
   use errorHandling
   
   use frametime_mod, only : getFrameTime ! mssllib
   use getFilterID_mod
   use omutility_mod      ! In mssllib various fits access routines
   use omphotometrydefs_mod
   use omconstants_mod    ! In mssllib defines various constants
   use omsourceparam_mod  ! In mssllib defines Struct SourceParameters
   use omparameters_mod 

   use omsrctable_mod     ! In mssllib for creating the source-list table
 
   use omgsrctable_mod     ! In mssllib for creating the source-list table
   use omsourcecelestial_mod
   use omprogress_mod
   use omsourceregion_mod      
   use omallocation_mod
   use omparameters_mod
   use omskybackground_mod
!   use omsourceparameter_mod

   use omgsourceparameter_mod
    
   implicit none


   private ! all, but:

   public :: myImage
   public :: OmSourceRegion      ! type definition
   public :: ObsDetails           ! NB
   public :: ImageWindow          !

   public  :: omsource_init       ! subroutine ...()
   public  :: omsource_finish     ! subroutine ...()
   public  :: omsource_loadsources  ! subroutine ... ()
   public  :: omsource_delete     ! subroutine ...()
   public  :: omsource_deleteList 

   public  :: omsource_replace  ! subroutine ...()

   public :: omsource_CreateGrismSourceList

   public :: omsource_CreateGrismSourceList1
   public :: omsource_CentroidSource 
   public :: omsource_UpdateSpectrumListFile
   public omsource_GetBackgroundMode 
   interface construct
      module procedure construct_od
   end interface

   interface destruct
      module procedure destruct_od
   end interface
 
   interface construct
      module procedure construct_iw
   end interface

   interface destruct
      module procedure destruct_iw
   end interface


   interface construct
      module procedure construct_celestial
   end interface

   interface destruct
      module procedure destruct_celestial
   end interface

   integer(int16) :: mySpectrumLength = 0
   integer(int16) :: mySpectrumSeparation = 0
   integer(int16) :: myZerothOrderSemiA = 0

   integer(int16) :: myFirstOrderSemiA = 0
   integer(int16) :: myZerothOrderSemiB = 0

   integer(int16) :: myFirstOrderSemiB= 0
         
  
         


contains
!*********************************************************************
! Info: Initialises the fortran part of omsource.
!       It loads the image pdata and properties, and -if any- it also
!       creates a source list from an already existing sources dataset
!*********************************************************************
subroutine omsource_init(sources, spectrumNo)

      type(GSourceParameters),  pointer    :: sources(:)
      integer, intent(inout) :: spectrumNo
 

      
!      type(OmProgressT)        :: reporter    ! ... of progress to GUI
      character(*) , parameter :: routine="omsource_old.omsource_init"
      integer(int32) :: ccdMode 
     real(single) :: sigma
      nullify(sources)       ! proper state
      

      ! --- initialise globals in this module







      myDeadFraction       = -1.
      myFrameTime          = -1.
      myDetectionRate      = -1.
      myDetectionMagnitude = -1.

      nullify(myImage)
      nullify(myQuality)

   
      ! --- extract the image from the file
      myImageSet = dataSet(stringParameter("imageset"), modify)

      ! load the image parameters (WINDOWX, etc) from the image fits file...
      call construct(myImageWindow, set = myImageSet)


      ! set the frame time etc...







      myFrameTime = myImageWindow%frameTime
      myDeadFraction = myImageWindow%deadFraction
      frameTime = myFrameTime
      deadFrac = myDeadFraction

!      call setProgress(reporter,4)



    ! Get the observation details for plotting
      call construct(myObsDetails, set = myImageSet)


   
      ccdMode = IMAGE
      call CAL_setState(OM, ccdModeId=ccdMode, filterId = filterIdToInt(myObsDetails%filter))
      !call setProgress(reporter,5)
      !TODO>

      select case(myObsDetails%filter)
          case ("GRISM1", "Grism1", "grism1")
              mySpectrumLength = 1100 /  myImageWindow%by
              mySpectrumSeparation = 900 / myImageWindow%by
              myFirstOrderSemiA = 375 !/  myImageWindow%by
              myFirstOrderSemiB = 5  !/  myImageWindow%by


              myZerothOrderSemiA = 10 !/  myImageWindow%by
              myZerothOrderSemiB = 5  ! /  myImageWindow%by

          case default
              mySpectrumLength = 600 / myImageWindow%by
              mySpectrumSeparation = 785 / myImageWindow%by
              myFirstOrderSemiA = 100 !/  myImageWindow%by
              myFirstOrderSemiB = 2  ! /  myImageWindow%by
              myZerothOrderSemiA = 5  !/  myImageWindow%by
              myZerothOrderSemiB = 2  !/  myImageWindow%by


      end select 


      ! create celestial information; also sets state of CAL
      call construct(myCelestial, set=myImageSet)
!      call setProgress(reporter, 6)



      !************************************* 
      ! Get the Plate scale (arcsecs/pixel)
      !*************************************
      call CAL_omGetPlateScale(0.55_double, myImageWindow%plateScale)
     
!write(*, *) "Platescale=",  myImageWindow%plateScale

  
!      call setProgress(reporter, 7)

      ! Get the RA and DEC of the image centre
!      call setLabel(reporter,"Computing sky coordinates")




!      call getCelestialCentre(myCelestial, myImageWindow%RA, myImageWindow%Dec)




!      call setProgress(reporter, 8)

      ! Get the pixels
      call createImageArray(myImageSet, myImage)

    ! Compute the background mode
      myBgrMode = backgroundMode(myImage, int(ubound(myImage, 1), int16), &
          int(ubound(myImage, 2), int16), 2.0, sigma) 
      myBgrModeSigma = max(2.0, sqrt(myBgrMode))










!      call setProgress(reporter, 9)

      ! Get the quality of the pixels
      call createQualityArray(myImageSet, myQuality)
!      call setProgress(reporter, 10)
!      call destruct(reporter)
      ! --- finished general information


      ! Load the old sourcelist file if one has been given on the
      ! command line
  
  



      call loadSources(sources, stringParameter("oldsrclistset"), spectrumNo)






!      if (associated(sources)) call setCelestial(myCelestial, sources)


      mySourcesAreLoaded = associated(sources)






     
   end subroutine omsource_init

!*******************************************************************************
! Info: Loads the sources from the source-list file produced by omdetect & ommag
!       The 'sources' argument is populated with the detected sources.
!*******************************************************************************
subroutine omsource_loadsources(sources, sourceListFile, spectrumNo)
    type(GSourceParameters),  pointer    :: sources(:)
    
    character(256), intent(in) :: sourceListFile;
    integer, intent(inout) :: spectrumNo




 
    call loadSources(sources, sourceListFile, spectrumNo)
     
     
    ! Compute the RA and Declination of each source

!      call setLabel(reporter,"Celestial properties")      
!    call setCelestial(myCelestial, sources)

end subroutine omsource_loadsources


!*************************************************************
! omsource_delete
!
!
! Info: deletes source=sources(id) from the source file; then all sources are
!       renumbered to make the IDs consitent.
!************************************************************/
subroutine omsource_delete(sources, id)
    type(GSourceParameters),  pointer    :: sources(:)
    integer,                  intent(in) :: id

    type(GSourceParameters),  pointer :: v(:)
    integer                           :: l, h  ! lbound/ubound
    integer                           :: idx  ! iterator
    integer                           :: off  ! 

      ! assume that sources(id) exists
    l = maxval(lbound(sources))
    h = maxval(ubound(sources))
    off = 0

    call new(v, low = l, high = h - 1)
    do idx = l, h
        if (idx .eq. id) then
            off = -1
            cycle
        end if
        v(idx + off) = sources(idx)
    end do

    call delete(sources)
    sources => v
    nullify(v)

end subroutine omsource_delete
!***********************************************
! Deletes the whole source list
!**********************************************
subroutine omsource_deleteList(sources)

    type(GSourceParameters),  pointer    :: sources(:)

    if (.not. associated(sources)) return
    call delete(sources)   
    nullify(sources)


    call new(sources, low = 1, high = 0)

end subroutine omsource_deleteList


!******************************************************
!
!*****************************************************
subroutine omsource_replace(id, source, sources)
    integer, intent(in) :: id
    type(SourceParameters), intent(in) :: source
    type(SourceParameters),  pointer    :: sources(:)
      
    sources(id) = source

end subroutine omsource_replace 
!******************************************************
! Info: saves sources, and frees memory
! used globals:
!   myCelestial
!   myImage
!   myImageSet
!   myImageWindow
!   myObsDetails
!   myQuality
!******************************************************
subroutine omsource_finish(sources)
    type(GSourceParameters),  pointer    :: sources(:)
    


    ! Add the sources to a table in the new source-list file
    call omsource_save(sources,&
                       stringParameter("newsrclistset"),&
                       stringParameter("newspectrumset"),&
 
                      myImageSet,&
                       myImageWindow)
    ! Close the image set, as we don't need it anymore.
    call release(myImageSet)

    ! Free memory for this module globals
    call destruct(myCelestial)
    call destruct(myImageWindow)
    call destruct(myObsDetails)
    if (associated(myImage))   deallocate(myImage)
    if (associated(myQuality)) deallocate(myQuality)

    ! And the sources themselves
    call delete(sources)
  
  
end subroutine omsource_finish
!**************************************************
! Gets the 2-d image from the fits file
!**************************************************
subroutine createImageArray(dataSet,image)
    type(dataSetT), intent(in) :: dataSet
    real(single),   pointer    :: image(:, :)

    type(ArrayT)           :: handle
    integer(int8), pointer :: int8ImageP(:, :)     ! Integer8 array pointer.
    integer(int32),pointer :: int32ImageP(:, :)    ! Integer32 array pointer.
    integer(int16),pointer :: int16ImageP(:, :)    ! Integer16 array pointer.
    real(single),  pointer :: real32ImageP(:, :)   ! Real32 array pointer.
    real(double),  pointer :: real64ImageP(:, :)   ! Real64 array pointer.
    integer                :: sz(2)
    integer                :: status

    nullify(image)
    handle = array(dataSet, "PRIMARY", read)

    sz = dimensions(handle)
    if (all(sz > 0)) then
        allocate( image(0:sz(1)-1,0:sz(2)-1),stat=status)
!        call assertOn(status/=0,"omsource_old.createImageArray.image")

        select case(arrayDataType(handle))
        case(integer8)
            int8ImageP => int8Array2Data(handle)
            image = real(int8ImageP)

        case(integer16)
            int16ImageP => int16Array2Data(handle)
            image = real(int16ImageP)

        case(integer32)
            int32ImageP => int32Array2Data(handle)
            image = real(int32ImageP)

        case(real32)
            real32ImageP => real32Array2Data(handle)
            image = real32ImageP

        case(real64)
            real64ImageP => real64Array2Data(handle)
            image = real64ImageP
        end select
    end if

    call release(handle)
end subroutine createImageArray
!*********************************************************
!
!*********************************************************
subroutine createQualityArray(dataSet,quality)
    type(dataSetT), intent(in) :: dataSet
    logical(bool),  pointer    :: quality(:, :)



    type(ArrayT)                :: handle
    integer(int8), pointer :: int8ArrayP(:, :) => null()
    integer(int16), pointer :: int16ArrayP(:, :) => null()

    integer(int32)              :: sz(2)
    integer                     :: status

    nullify(quality)
    handle = array(dataSet, "QUALITY", read)

    sz = dimensions(handle)

    allocate( quality(0 : sz(1) - 1, 0 : sz(2) - 1), stat=status)


    if (all(sz > 0)) then
        


        select case(arrayDataType(handle))
            case(integer8)            
                int8ArrayP => int8Array2Data(handle)

                where (int8ArrayP .ne. 0)
                    quality = .true.
                elsewhere
                    quality = .false.
                end where
            case(integer16)
                int16ArrayP => int16Array2Data(handle)

                where (int16ArrayP .ne. 0)
                    quality = .true.
                elsewhere
                    quality = .false.
                end where
       
          end select
      
  
    end if

    call release(handle)







end subroutine createQualityArray


!*********************************************************************************
! subroutine loadSources
! Info: loads sources found in dataset(setName), and returns a newly created vector
!       of sources.
! 
!********************************************************************************
subroutine loadSources(sources, setName, spectrumNo)

    type(GSourceParameters), pointer    :: sources(:)
    character(*), intent(in) :: setName
    integer, intent(inout) :: spectrumNo
  
    ! locals
    integer(int32)        :: nSources
    integer :: i
    type(DataSetT)        ::  set
    type(TableT)          :: inTableH   
    type(OmProgressT)     :: reporter    ! ... of progress to GUI    
   



      ! --- don't bother if none available
    if (trim(setName) == "" .or. len(trim(setName)) .lt. 2 ) then
       

        return
    end if
    ! Read in the OM OSW Source List file.
    set = dataSet(trim(setName), read)



    myExposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  







    if (.not. hasBlock(set, "SRCLIST")) then
      


 

        ! Incompatible file - can't continue
        write (errstr, *) "Could not find any compatibile blocks in set ", trim(setName)
        call fatal("BadFileType")

    else




        ! Get a handle on the source list binary table extension.
        inTableH = table(set, "SRCLIST")

    end if




    

    ! Get the limiting magnitude of the image
    myDetectionRate      = GetReal32Attribute(set, "CRLIMIT")
    myDetectionMagnitude = GetReal32Attribute(set, "MAGLIMIT")
    
  

    call construct(reporter, label="Loading sourcelist data", steps = 23)


    ! Get the number of rows in the source list binary table extension.








    nSources = numberOfRows(inTableH)




    if (nSources == 0) then
        call destruct(reporter)

        call release(inTableH)
        call release(set)
        return

    end if

    call delete(sources)
    call new(sources, low = 1, high = nSources)
 
    ! Get the data for each source from the source list file
    call setProgress(reporter, 1)
  
    call LoadReal32Column(inTableH, "XPOS",  sources%xPos)
    call setProgress(reporter, 2)
  
    call LoadReal32Column(inTableH, "YPOS",  sources%yPos)
    call setProgress(reporter, 3)
  
    call LoadReal32Column(inTableH, "FWHM_MAJ",  sources%semiA)
    call setProgress(reporter, 4)
  
    call LoadReal32Column(inTableH, "FWHM_MIN",  sources%semiB)
    call setProgress(reporter, 5)
  
    call LoadReal32Column(inTableH, "PA", sources%theta)
    call setProgress(reporter, 6)
  
  

    call LoadInt8Column(inTableH, "EFLAG", sources%extensionFlag)
    
    call setProgress(reporter, 19)

    ! Get the RAS and decs if they exist
    call LoadReal64Column(inTableH, "RA", sources%rightAscension)
    call setProgress(reporter, 20)
  
    call LoadReal64Column(inTableH, "DEC", sources%declination)
    call setProgress(reporter, 21)
  
    call LoadInt8Column(inTableH, "EFLAG", sources%extensionFlag)



    if (.not. hasColumn(inTableH, "SPECTR_ID")) then

 
       call fatal("no column", "SPECTR_ID")
    end if

    call LoadInt32Column(inTableH, "SPECTR_ID", sources%spectrumNo)

    spectrumNo = 0
    do i = 1, nSources
        if (sources(i)%spectrumNo > 0 .and. sources(i)%spectrumNo < 10000) spectrumNo = spectrumNo + 1         
    end do

    call LoadInt32Column(inTableH, "REL2SRCNUM", sources%origin)


    call setProgress(reporter,22)

    call release(inTableH)
    call release(set)


    ! Convert width to arcsecs
  
    sources%semiA = sources%semiA / ( myImageWindow%bx *  myImageWindow%plateScale * 2.354)
    sources%semiB = sources%semiB / ( myImageWindow%bx *  myImageWindow%plateScale * 2.354)
         
    call setProgress(reporter,23)


    call destruct(reporter)


end subroutine loadSources

!**********************************************************
! Info: saves the sources into dataset(fileName)
! Used globals:
!   mySourcesAreLoaded
!   myDetectionRate
!   myDetectionMagnitude
!   myFrameTime
!   myDeadFraction
!   myNullValue
! TODO: save unbinnend radius? save baground radii?
!*********************************************************
subroutine omsource_save(sources, sourceListFile, spectrumFile, imageSet, iw)
    type(GSourceParameters), pointer    :: sources(:)
    character(*),            intent(in) :: sourceListFile, spectrumFile
    type(DataSetT),          intent(in) :: imageSet
    type(ImageWindow),       intent(in) :: iw
!    type(ObsDEtails),        intent(in) :: od

    integer                :: nSources

    type(OmProgressT)      :: reporter    ! ... of progress to GUI
    character(*) , parameter :: routine="omsource_old.omsource_save"

    if (.not.associated(sources)) return ! don't bother
    nSources = size(sources)


    ! notify the GUI that we are busy...
    call construct(reporter, label="Saving source list- please wait", steps = 2)
    call setProgress(reporter, 1)


    ! Create the grism source-list table
    call CreateGSourceListFile(sourceListFile,  &
                               spectrumFile,    &
                              imageSet,         &
                              nSources,         &
                              sources,          &
                              iw%plateScale,    &
                              myBgrMode,        &
                              3.0,              &
                              myExposure,       & 
                              iw%x0,            &
                              iw%y0,            & 
                              iw%width,         &
                              iw%height) 

    call setProgress(reporter, 2)

    call destruct(reporter)

end subroutine omsource_save

!****************************************************************
! Info: computes all properties of the sources specified
! Note: * Each input source is 'empty' save the (xCoord,yCoord) properties
!       * 'trimage' argument is a residual, and should be removed
!****************************************************************
subroutine computeParameters(image, sources)
    real(single),           pointer    :: image(:, :)
    type(SourceParameters),pointer    :: sources(:)
    integer(int32)                     :: i,nSources,nx, ny
    integer(int32)                     :: i0, j0
    integer(int16) :: sourceNumber = 1
   
    nx = size(image, 1)
    ny = size(image, 2)
    nSources = size(sources)

    

    ! Compute the centre coordinates of the new sources
    do i = 1, nSources
       
        call computeCentreOfSource(nx, ny, image, &
              sources(i)%xPos, sources(i)%yPos, &
              i0, j0)
        sources(i)%iMin = max(i0 - 3, 0)
        sources(i)%iMax = min(i0 + 3, nx - 1)
        sources(i)%jMin = max(j0 - 3, 0)
        sources(i)%jMax = min(j0 + 3, ny - 1)

        call ComputePointSourceParameters(sourceNumber, nx - 1, ny - 1, image, &
                i0, j0, sources(i))

   
    end do
    

    ! Compute the RA and Declination of each source
    call setCelestial(myCelestial, sources)
end subroutine computeParameters


!************************************************************

   ! Info: find a more accurate centre pixel for (xPos,yPos)
!***********************************************************
subroutine computeCentreOfSource(nx, ny, image, xPos, yPos, i0, j0)
    integer(int32), intent(in) :: nx, ny
    real(single), intent(in) :: image(1 : nx, 1 : ny)
     
    real(single), intent(in) :: xPos, yPos

    integer(int32), intent(out) :: i0, j0
    integer(int32) :: il, iu, jl, ju

    integer(int32) :: i, j
    real(single) :: maximum

      ! current pixel centre

    i0 = int(xPos - 0.5)
    j0 = int(yPos - 0.5)

    ! now find a more accurate centre for the image
    il = i0 - 5
    il = max (il, 0)
    iu = i0 + 5
    iu = min (iu, nx - 1)

    jl = j0 - 5
    jl = max (jl, 0)
    ju = j0 + 5
    ju = min (ju, ny - 1)
    maximum = 1e-6

    do i = il, iu
        do j = jl, ju
            if (image(i, j) .gt. maximum) then
                maximum = image(i, j)
                i0 = i
                j0 = j
            end if
        end do
    end do

end subroutine computeCentreOfSource
!***************************************
!
!***************************************
subroutine construct_od(o, set)
    type(ObsDetails),  intent(out)    :: o
    type(DataSetT),    intent(in)     :: set
      
    call GetStringAttribute(set, "TELESCOP", o%telescope)
    call GetStringAttribute(set, "OBSERVER", o%observer)
    call GetStringAttribute(set, "OBJECT",   o%object)
    call GetStringAttribute(set, "FILTER",   o%filter)
    call GetStringAttribute(set, "INSTRUME", o%instrument)
    call GetStringAttribute(set, "OBS_ID",   o%obsID)
    call GetStringAttribute(set, "EXP_ID",   o%expID)
    call GetStringAttribute(set, "DATE-OBS", o%startObs)
    call GetStringAttribute(set, "DATE-END", o%endObs)
    o%exposure = GetReal32Attribute(set, "EXPOSURE")
end subroutine construct_od
!***************************************
!
!***************************************
subroutine destruct_od(o)
    type(ObsDetails), intent(inout)    :: o
    o%telescope=''
    o%exposure=0.
end subroutine destruct_od
!***************************************
!
!***************************************
subroutine construct_iw(w, set)
    type(ImageWindow), intent(out)    :: w
    type(DataSetT),    intent(in)     :: set
      
    type(TableT)                      :: t
    type(ColumnT)                     :: c
    integer(int32), pointer            :: ptr(:) ! release on table release
    type(DataSetT) :: wdxDataSet      


    t = table(set, "MODES")
    c = column(t, "WINDOWX0", read)
    ptr => int32Data(c)
    w%x0  = ptr(0)
    c = column(t, "WINDOWY0", read)
    ptr => int32Data(c)
    w%y0  = ptr(0)
    c = column(t, "WINDOWDX", read)
    ptr => int32Data(c)
    w%width  = ptr(0)
    c = column(t, "WINDOWDY", read)
    ptr => int32Data(c)
    w%height  = ptr(0)
    myExposure = GetReal32Attribute(set, "EXPOSURE", 0.0)  

   
    call release(t)

    ! compute binning
    w%binBpe = booleanAttribute(set, "BINBPE")
    w%bx = 2 ** int32Attribute(set, "BINAX1")
   
    w%by = 2 ** int32Attribute(set, "BINAX2")
   
    
    if (w%binBpe) then
        w%bx  = 2 
        w%by  = 2 
    end if
    binX = w%bx
    binY = w%by


   w%filter  = stringAttribute(set, "FILTER")
   if (trim(w%filter) .ne. "GRISM1" .and. trim(w%filter) .ne. "grism1" .and. &
       trim(w%filter) .ne. "GRISM2" .and. trim(w%filter) .ne. "grism2") then
        call fatal("Program only works with grism data")
    end if


    w%datamode= stringAttribute(set,"DATAMODE")
   

    ! CCD frame-time and dead-fraction
    
    if (.not. hasAttribute(set,  "FRAMTIME")) then
        wdxDataSet = dataSet(stringParameter("wdxset"), read)


        call getFrameTime(wdxDataSet, w%frameTime, w%deadFraction, w%binBpe)
        call release(wdxDataSet)
        call setAttribute(set, "FRAMTIME", w%FrameTime, "millisecs", "CCD frame time")
        call setAttribute(set, "DEADFRAC", w%DeadFraction, "", "CCD dead fraction")
        w%frameTime = w%frameTime / 1000.0
       
    else     
        w%frameTime = real64Attribute(set, "FRAMTIME") / 1000.0
        w%deadFraction = real64Attribute(set, "DEADFRAC")

    end if 
end subroutine construct_iw
!***************************************
!
!***************************************
subroutine destruct_iw(w)
    type(ImageWindow), intent(inout)    :: w
    w%filter=''
    w%datamode=''
end subroutine destruct_iw

!*******************************************************
!
! Creates a source-list for input to omgrism
!*******************************************************
subroutine omsource_CreateGrismSourceList(n, yZero, xFirst, &
            x1, x2, y1, y2, theSources, oldSources, newSources, &
            spectrumNo, centroidyZero, centroidxFirst)
    integer, intent(in) :: n

    real(single), intent(in) :: yZero(1 : n), xFirst(1 : n)
    real(single), intent(in) :: x1(1 : n), x2(1 : n)
    real(single), intent(in) :: y1(1 : n), y2(1 : n)

    real(single) :: x0, y0
    type(GSourceParameters), pointer    :: theSources(:)  ! The global source-list
 
    type(GSourceParameters), pointer    :: oldSources(:)
    type(GSourceParameters), pointer :: newSources(:)
    integer(int32), intent(inout) :: spectrumNo

    integer(int8) :: centroidyZero, centroidxFirst
    integer(int16) :: number = 1

    integer i, j, k, i0, j0

    type (SourceParameters), pointer :: tempSource(:) => null()
     
    integer(int16), pointer :: levelImage(:, :) => null() 
    integer(int16) :: sourceNo = 0
  
    integer nx, ny
    integer(int32) :: noSources
    type(DataSetT) :: dataSetH
    type(TableT)   :: tableH


   
    sourceNo = 0    
    if (associated(theSources)) then
        sourceNo = size(theSources)
    end if 
    if (associated(oldSources)) then 
    
        sourceNo = sourceNo + size(oldSources)
        
    end if
   
    k = 0
    
    nx = ubound(myImage, 1)
    ny = ubound(myImage, 2)

      
  
    levelImage => Allocate2dInt16(0, nx, 0, ny, "levelImage")

    k = 1
    ! Loop through the new sources
    noSources = 0
    
    call new(newSources, low = 1, high = 2 * n)

    allocate(tempSource(1 : 2 * n))
  
    do i = 1, n
       
        !**************************************************************
        ! Find the centroided position of the zeroth order point-source
        !**************************************************************
        noSources = noSources + 1
        sourceNo = sourceNo + 1    
    
        
        x0 = (x1(i) + x2(i)) / 2.0
        y0 = y1(i)
        i0 = max(0, nint(x0))
        j0 = max(0, nint(y0)) 
        i0 = min(nx, i0)
        j0 = min(ny, j0) 

        tempSource(noSources)%iMin = max(0, i0 - 20)
        tempSource(noSources)%iMax = min(nx, i0 + 20)
       
        tempSource(noSources)%jMin = max(j0 - 50, 0)
        tempSource(noSources)%jMax = min(j0 + 50, ny)
         
        tempSource%number = sourceNo

        !******************************************
        ! See if can do the centroiding
        !****************************************** 
        do j = 1, 3
            levelImage = 0
            where(myImage .ge. myBgrMode + myBgrModeSigma * (5.0 - real(j))) levelImage = number     




            if (sum(levelImage(tempSource(noSources)%iMin : &
                           tempSource(noSources)%iMax,  tempSource(noSources)%jMin : &
                           tempSource(noSources)%jMax)) .gt. 20) then


     
                call ComputeExtendedSourceParameters(number, nx, ny, myImage, &
                    levelImage, tempSource(noSources))
                exit  
            else
                if (j .eq. 3) then
                    tempSource(noSources)%xPos = xFirst(i)
                    tempSource(noSources)%yPos = yZero(i)
                end if
            end if
        end do
            




        newSources(noSources)%xPos = tempSource(noSources)%xPos




        if (centroidyZero .eq. 1) then ! User wishes to use centroided y-zero coordinate            

            newSources(noSources)%yPos = tempSource(noSources)%yPos


        else
            newSources(noSources)%yPos = yZero(i)
            tempSource(noSources)%yPos = yZero(i)

        end if

 
        newsources(noSources)%extensionFlag = 1
        tempSource(noSources)%extensionFlag = 1


        tempSource(noSources)%semiA =  real(myZerothOrderSemiA)
        tempSource(noSources)%semiB =  real(myZerothOrderSemiB)
     
        tempSource(noSources)%theta = 90.0
        newSources(noSources)%semiA = tempSource(noSources)%semiA
        newSources(noSources)%semiB = tempSource(noSources)%semiB


        newSources(noSources)%theta = tempSource(noSources)%theta
        newSources(noSources)%noSourcePixels = tempSource(noSources)%noSourcePixels


        spectrumNo = spectrumNo + 1

        newSources(noSources)%spectrumNo = sourceNo
   
        newSources(noSources)%origin = sourceNo + 1

        
  
        ! The first-order spectrum
        ! NB For visible the separation bewteen zero and first order is 785 / bin pixels
        ! and the length of the spectrum is 600 / bin pixels
        ! NB For visible the separation bewteen zero and first order is 900 / bin pixels
        ! and the length of the spectrum is 1100 / bin pixels
        j0 = nint(tempSource(noSources)%yPos - 0.5)

        j0 = min(ny, j0 + mySpectrumSeparation)


        noSources = noSources + 1
        sourceNo = sourceNo + 1


       
        tempSource(noSources)%iMin = max(0, i0 - 10)
        tempSource(noSources)%iMax = min(nx, i0 + 10)
       

        tempSource(noSources)%jMin = max(0, j0 - mySpectrumLength / 2)
        tempSource(noSources)%jMax = min(ny, j0 + mySpectrumLength / 2)

         


        do j = 1, 3
            levelImage = 0
            where(myImage .ge. myBgrMode + myBgrModeSigma * (5.0 - real(j))) levelImage = number     




            if (sum(levelImage(tempSource(noSources)%iMin : &
                           tempSource(noSources)%iMax,  tempSource(noSources)%jMin : &
                           tempSource(noSources)%jMax)) .gt. 20) then


     
                call ComputeExtendedSourceParameters(number, nx, ny, myImage, &
                    levelImage, tempSource(noSources))
                exit
            else
                if (j .eq. 3) then
                    tempSource(noSources)%xPos = xFirst(i)
                    tempSource(noSources)%yPos = y2(i)
                end if
            end if
        end do
          

          
 

        
        if (centroidxFirst .eq. 1) then ! User using the centroided x position            
        
            newSources(noSources)%xPos = tempSource(noSources)%xPos
            newSources(noSources - 1)%xPos = tempSource(noSources)%xPos
 
        else

            newSources(noSources - 1)%xPos = xFirst(i)  ! Added Friday the 13th August
            tempSource(noSources - 1)%xPos = xFirst(i)  ! Added Friday the 13th August

            newSources(noSources)%xPos = xFirst(i)
            tempSource(noSources)%xPos = xFirst(i)


        end if

 

        newsources(noSources)%extensionFlag = 1
        tempSource(noSources)%extensionFlag = 1

     

        tempSource(noSources)%semiA =  real(myFirstOrderSemiA)
        tempSource(noSources)%semiB =  real(myFirstOrderSemiB) 
        tempSource(noSources)%theta = 90.0
     

        newSources(noSources)%yPos = tempSource(noSources)%yPos
        newSources(noSources)%semiA = tempSource(noSources)%semiA
        newSources(noSources)%semiB = tempSource(noSources)%semiB
        newSources(noSources)%theta = tempSource(noSources)%theta
        newSources(noSources)%noSourcePixels = tempSource(noSources)%noSourcePixels

        newSources(noSources)%spectrumNo = -1
        newSources(noSources)%origin = -1
              
    end do

    deallocate(levelImage)
    nullify(levelImage)



    !*****************************************************
    ! Append the new sources to the old source list
    !****************************************************

    ! Create the source-list table for omgrism


    call CreateSourceListFile(trim(stringParameter("newsrclistset")), &
                              myImageSet,                  &
                              noSources,                   &
                              tempSource,                  &
                              myimagewindow%plateScale,    &
                              12.0,                        & 
                              myBgrMode,                   &
                              3.0,                         &
                              myExposure,                  & 
                              myimageWindow%x0,            &
                              myimageWindow%y0,            & 
                              myimageWindow%width,         &
                              myimageWindow%height) 


    dataSetH = dataSet(trim(stringParameter("newsrclistset")), modify)
    tableH = Table(dataSetH, "SRCLIST") 

    call setAttribute(tableH, "FILENAME", trim(stringParameter("newsrclistset")), "Name of file containing the spectra")
 
    call release(tableH)
    call release(dataSetH)


    !*****************************************************
    ! Append the new sources to the old source list
    !****************************************************
    call AppendToSourceList(oldSources, newSources)

 
    deallocate(tempSource)
    nullify(tempSource)
     

end subroutine omsource_CreateGrismSourceList



!*******************************************************
!
! Creates a source-list for input to omgrism
!*******************************************************
subroutine omsource_CreateGrismSourceList1(xZero, yZero, &
            theSources, oldSources, newSources, &
            spectrumNo)
   
    real(single), intent(in) :: xZero, yZero
  

   
    type(GSourceParameters), pointer    :: theSources(:)  ! The global source-list
 
    type(GSourceParameters), pointer    :: oldSources(:)
    type(GSourceParameters), pointer :: newSources(:)
    integer(int32), intent(inout) :: spectrumNo

  


    type (SourceParameters), pointer :: tempSource(:) => null()
     
 
    integer(int16) :: sourceNo = 0
  
 
    integer(int32) :: noSources
    type(DataSetT) :: dataSetH
    type(TableT)   :: tableH


!write(*, *) "creategrismsourcelist1 line 1182 xZero=", xZero, " yzero=", yZero   
    sourceNo = 0    
    if (associated(theSources)) then
        sourceNo = size(theSources)
    end if 
 
   if (associated(oldSources)) then 
    
        sourceNo = sourceNo + size(oldSources)
        
    end if
   
  
!    write(*, *) "creategrismsourcelist1 line 1195"   

 

      

 
    ! Loop through the new sources
    noSources = 0
    
    call new(newSources, low = 1, high = 2)






!write(*, *) "creategrismsourcelist1 line 1212"   



    allocate(tempSource(1 : 2))
  
   
       
    !**************************************************************
    ! Th zeroth order point-source
    !**************************************************************
    noSources = noSources + 1
    sourceNo = sourceNo + 1    
    
!        write(*, *) "creategrismsourcelist1 line 1226"   

  
         
    tempSource%number = sourceNo

    newSources(noSources)%xPos = xZero
    newSources(noSources)%yPos = yZero
               
          
  
    tempSource(noSources)%xPos = xZero
    tempSource(noSources)%yPos = yZero
               
        

    newsources(noSources)%extensionFlag = 1
    tempSource(noSources)%extensionFlag = 1


    tempSource(noSources)%semiA =  real(myZerothOrderSemiA)
    tempSource(noSources)%semiB =  real(myZerothOrderSemiB)
     
    tempSource(noSources)%theta = 90.0
    newSources(noSources)%semiA = tempSource(noSources)%semiA
    newSources(noSources)%semiB = tempSource(noSources)%semiB


    newSources(noSources)%theta = tempSource(noSources)%theta
    newSources(noSources)%noSourcePixels = tempSource(noSources)%noSourcePixels


    spectrumNo = spectrumNo + 1

    newSources(noSources)%spectrumNo = sourceNo
   
    newSources(noSources)%origin = sourceNo + 1

        
  



!write(*, *) "creategrismsourcelist1 line 1280"   


    ! The first-order spectrum
    ! NB For visible the separation bewteen zero and first order is 785 / bin pixels
    ! and the length of the spectrum is 600 / bin pixels
    ! NB For visible the separation bewteen zero and first order is 900 / bin pixels
    ! and the length of the spectrum is 1100 / bin pixels
    


    noSources = noSources + 1
    sourceNo = sourceNo + 1


       
    tempSource(noSources)%xPos = xZero
    tempSource(noSources)%yPos = yZero + real(mySpectrumSeparation) 
             
          

          
    newSources(noSources)%xPos = xZero
    newSources(noSources)%yPos = yZero + real(mySpectrumSeparation) 

       


    newsources(noSources)%extensionFlag = 1
    tempSource(noSources)%extensionFlag = 1

     

    tempSource(noSources)%semiA =  real(myFirstOrderSemiA)
    tempSource(noSources)%semiB =  real(myFirstOrderSemiB) 
    tempSource(noSources)%theta = 90.0
     

    newSources(noSources)%semiA = tempSource(noSources)%semiA
    newSources(noSources)%semiB = tempSource(noSources)%semiB
    newSources(noSources)%theta = tempSource(noSources)%theta
    newSources(noSources)%noSourcePixels = tempSource(noSources)%noSourcePixels

    newSources(noSources)%spectrumNo = -1
    newSources(noSources)%origin = -1
             

  
!write(*, *) "creategrismsourcelist1 line 1345"   



    !*****************************************************
    ! Append the new sources to the old source list
    !****************************************************

    ! Create the source-list table for omgrism

!write(*, *) "source 1 x=", tempSource(1)%xpos, " y=", tempSource(1)%ypos
!write(*, *) "source 2 x=", tempSource(2)%xpos, " y=", tempSource(2)%ypos



    call CreateSourceListFile(trim(stringParameter("newsrclistset")), &
                              myImageSet,                  &
                              noSources,                   &
                              tempSource,                  &
                              myimagewindow%plateScale,    &
                              12.0,                        & 
                              myBgrMode,                   &
                              3.0,                         &
                              myExposure,                  & 
                              myimageWindow%x0,            &
                              myimageWindow%y0,            & 
                              myimageWindow%width,         &
                              myimageWindow%height) 


    dataSetH = dataSet(trim(stringParameter("newsrclistset")), modify)
    tableH = Table(dataSetH, "SRCLIST") 

    call setAttribute(tableH, "FILENAME", trim(stringParameter("newsrclistset")), "Name of file containing the spectra")
 
    call release(tableH)
    call release(dataSetH)




!write(*, *) "creategrismsourcelist1 line 1382"   



    !*****************************************************
    ! Append the new sources to the old source list
    !****************************************************
    call AppendToSourceList(oldSources, newSources)

 

!write(*, *) "creategrismsourcelist1 line 1393"   




    deallocate(tempSource)
    nullify(tempSource)
     

end subroutine omsource_CreateGrismSourceList1





!*****************************************************
! Appends newSources list  to oldSources list
!****************************************************

subroutine AppendToSourceList(oldSources, newSources)

    type(GSourceParameters), pointer :: oldSources(:)  
    type(GSourceParameters), pointer :: newSources(:)
    type(GSourceParameters), pointer :: tempSources(:)

    integer(int32) :: k

    if (.not. associated(newSources)) return; 


    if (.not. associated(oldSources)) then

        call new (oldSources, low = 1, high = size(newSources))

        oldSources = newSources

        return
    end if

    k = size(oldSources)
 
    call new(tempSources, low = 1, high = k)
 
    tempSources = oldSources
    call delete(oldSources)
    nullify(oldSources)  
    call new(oldSources, low = 1, high = k + size(newSources))
 
    oldSources(1 : k) = tempSources
    call delete(tempSources)
    nullify(tempSources)    

    oldSources(k + 1 : k + size(newSources)) = newSources

    
end subroutine AppendToSourceList




!*************************************************
!
! Returns the background mode
!*************************************************
real(single) Function omsource_GetBackgroundMode()
    omsource_GetBackgroundMode = myBgrMode
    
end function omsource_GetBackgroundMode
!************************************************************
!
!************************************************************
subroutine omsource_CentroidSource(xl, xr, yl, yu, &
                 xPos, yPos)
    integer, intent(in) :: xl, xr, yl, yu
    real(single), intent(out) :: xPos, yPos
    integer(int16) :: sourceNo = 0
    type (SourceParameters) :: source   
    integer(int16), pointer :: levelImage(:, :) => null() 
  
    
    levelImage => Allocate2dInt16(0, ubound(myImage, 1), 0, ubound(myImage, 2), "levelImage")

       

    source%iMin = max(0, xl)
    source%iMax = min(xr, size(myImage, 1) - 1)
    source%jMin = max(0, yl)
    source%jMax = min(yu, size(myImage, 2) - 1)
    where(myImage .ge. max(2.0, myBgrMode + 3.0 * sqrt(myBgrMode))) levelImage = sourceNo

    call ComputeExtendedSourceParameters(sourceNo, size(myImage, 1), size(myImage, 2), &
                myImage, levelImage, source)
     

    deallocate(levelImage)
    nullify(levelImage)

    xPos = source%xpos
    yPos = source%ypos 
    
end subroutine omsource_CentroidSource

!************************************************************
!
! Subroutine updatespectrumlistfiles
! Updates the spectrum file by appending the data in 
! the spectrum file created by omgrism
! to it
!
!*************************************************************
subroutine omsource_UpdateSpectrumListFile(myNewSources)
    type(GSourceParameters), pointer :: myNewSources(:)
  
    type(DataSetT) :: oldSpectrumSet
    type(DataSetT) :: newSpectrumSet
   
    type(TableT)   :: tab1, tab2
    
    type(ColumnT)  :: column1

    integer :: i, j
    
    character(4)  :: rate
    character(7) :: tableName
    integer :: blockNo
    




    

    ! The spectrum list file created by omgrism
    newSpectrumSet = dataSet(trim(stringParameter("newspectrumset")), MODIFY)

    rate = "RATE"


    if (.not. setExists(trim(stringParameter("oldspectrumset")))) then


        
        ! The spectrum list file created by omgrism
        oldSpectrumSet = dataSet(trim(stringParameter("oldspectrumset")), create)
    
        call copyAllAttributes( attributable(newSpectrumSet), attributable(oldSpectrumSet) )

        tab1 = table(newSpectrumSet,  0, READ )


        tab2 = addTable(oldSpectrumSet, "RATE001", numberOfRows(tab1))

        call copyAllAttributes(attributable(tab1), attributable(tab2))






        do j = 0, numberOfColumns(tab1) - 1
            column1 = Column(tab1, j)

            call copyColumn(tab2, column1, name(column1))

            
        end do


        call release(oldSpectrumSet)
        call release(newSpectrumSet)


        return   
    end if   


    oldSpectrumSet = dataSet(trim(stringParameter("oldspectrumset")), MODIFY)
   
    !***************************************************************
    ! Loop through the new source-list table and see which one
    ! has a spectum. For every one that has copy the
    ! block from the new spectrum file to the old one,
    ! with an ammended block name
    !***************************************************************

    blockNo = -1



    do i = 1, size(myNewSources)
        if (myNewSources(i)%spectrumNo < 0 .or. myNewSources(i)%spectrumNo > 10000) cycle  ! Not a spectrum
       
      
        blockNo = blockNo + 1
        write(tableName, '(a4, i3.3)') rate,  ((i - 1) * 2 + 1)
        
        tab1 = table(newSpectrumSet,  blockNo, READ )

        write(tableName, '(a4, i3.3)') rate,  myNewSources(i)%spectrumNo
      


        tab2 = addTable(oldSpectrumSet, trim(tableName), numberOfRows(tab1))

        call copyAllAttributes(attributable(tab1), attributable(tab2))

        do j = 0, numberOfColumns(tab1) - 1
            column1 = Column(tab1, j)

            call copyColumn(tab2, column1, name(column1))

        end do
    end do








    call release(oldSpectrumSet)
    call release(newSpectrumSet)

    call delete( myNewSources )
    nullify( myNewSources )

end subroutine omsource_UpdateSpectrumListFile

end module omsource_old










